/*
 * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include "fu-usb-device.h"

G_BEGIN_DECLS

const gchar	*fu_usb_device_get_platform_id		(FuUsbDevice	*self);

G_END_DECLS
